/* 
 * EMBL - European Bioinformatics institute
 * MSD Group
 * VAMSAS Project
 * 
 *Created on 28 oct. 2005
 * @author <a href="mailto:pierre@ebi.ac.uk">Pierre MARGUERITE</a>
 * @version  1  
 * @since 28 oct. 2005
 */
package uk.ac.ebi.msd.vamsas.vamav.documentmanagement;


import uk.ac.ebi.msd.vamsas.vamav.session.VAMAVSession;
import uk.ac.ebi.msd.vamsas.vamav.util.data.DataBuffer;
import uk.ac.ebi.msd.vamsas.vamav.util.data.PDBEntry;
import uk.ac.ebi.msd.vamsas.vamav.util.file.PlainTextFileReader;
import uk.ac.ebi.msd.vamsas.vamav.util.logger.VAMAVLogger;
import uk.ac.ebi.msd.vamsas.vamav.util.logger.VAMAVLoggerImpl;
import uk.ac.ebi.msd.vamsas.vamav.util.pool.DataBufferPool;
import uk.ac.ebi.msd.vamsas.vamav.util.pool.StringBufferPool;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;

/**
 * Convenient methods for PDB entry manipulation
 * 
 * @author <a href="mailto:pierre@ebi.ac.uk">Pierre MARGUERITE</a>
 * @version 1
 * @since 28 oct. 2005 
 * 
 */
public class PDBUtil
	{
		private  VAMAVLogger logger = new VAMAVLoggerImpl(this.getClass());
		
		private static Hashtable<String, String> AminoAcids = null; 
		
		/**
		 * Constructor 
		 *
		 */
		public PDBUtil(String session)
			{
				super();
				logger.setSession(session);
				
			}

		/**
		 * Constructor 
		 *
		 */
		public PDBUtil(VAMAVSession session)
			{
				super();
				logger.setSession(session);
				
			}
		/**
		 * PDB Format 
		 * http://www.rcsb.org/pdb/docs/format/pdbguide2.2/part_35.html
		 * 
		 * @param entry
		 * @return
		 */
		public  ArrayList<Chain> GetChain(PDBEntry entry)
		{
			if( entry == null) return null;
			File pdbFile = entry.getPdbFile();
			if (pdbFile == null || !pdbFile.exists() )
				{
					return null;
				}
			logger.sessionInfo("Retrieving PDB sequence from PDB file.");
			ArrayList<Chain>  chains = null;
			PlainTextFileReader reader = new PlainTextFileReader(pdbFile);
			try
				{
					reader.openFile();
					DataBuffer line = null;
					Hashtable<String, StringBuffer> retrievingChains = new  Hashtable<String, StringBuffer>();
					while (( line = reader.getLine()) !=null)
						{
							/*System.out.println("line "+line);*/
					if( line.toString().startsWith("SEQRES"))//.matches("^SEQRES*"))
						{//chain line found
							/*System.out.println("SEQRES found ");
							System.out.println("line "+line);*/
							//12             Character       chainID
							char chainID = line.charAt(12 -1);
							String chainName = String.valueOf(chainID);
							/*System.out.println("chainName "+chainName);*/
							StringBuffer seq = retrievingChains.get(chainName.intern());
							if (seq == null)
								{
									seq = StringBufferPool.getString();
								}
							
							/*	System.out.println("chain retrieves.");*/
//							20 23        Residue name    resName       Residue name.
							if(line.length()>=23 )
								{
									String aa = line.substring(20-1, 23-1);
									if( aa != null)
										{
											aa = aa.trim();
											if(aa.length()>0)
												seq.append(get1LetterCodeFrom3LettersCode(aa));
											
										}		
								}
							//24 - 27        Residue name    resName       Residue name.
							if(line.length()>=27 )
								{
									String aa = line.substring(24-1, 27-1);
									if( aa != null)
										{
											aa = aa.trim();
											if( aa.length()>0)
												seq.append(get1LetterCodeFrom3LettersCode(aa));
										}
								}
							//28 - 30        Residue name    resName       Residue name.
							if(line.length()>=31 )
								{
									String aa = line.substring(28-1, 31-1);
									if( aa != null)
										{
											aa = aa.trim();
											if( aa.length()>0)
												seq.append(get1LetterCodeFrom3LettersCode(aa));
										}
								}
							//32 - 34        Residue name    resName       Residue name.
							if(line.length()>=35 )
								{
									String aa = line.substring(32-1, 35-1);
									if( aa != null)
										{
											aa = aa.trim();
											if( aa.length()>0)
												seq.append(get1LetterCodeFrom3LettersCode(aa));
										}
								}
							//36 - 38        Residue name    resName       Residue name.
							if(line.length()>=39 )
								{
									String aa = line.substring(36-1, 39-1);
									if( aa != null)
										{
											aa = aa.trim();
											if( aa.length()>0)
												seq.append(get1LetterCodeFrom3LettersCode(aa));
										}
								}
							//40 - 42        Residue name    resName       Residue name.
							if(line.length()>=42 )
								{
									String aa = line.substring(40-1, 43-1);
									if( aa != null)
										{
											aa = aa.trim();
											if( aa.length()>0)
												seq.append(get1LetterCodeFrom3LettersCode(aa));
										}
								}
							//44 - 46        Residue name    resName       Residue name.
							if(line.length()>=47 )
								{
									String aa = line.substring(44-1, 47-1);
									if( aa != null)
										{
											aa = aa.trim();
											if( aa.length()>0)
												seq.append(get1LetterCodeFrom3LettersCode(aa));
										}
								}
							//48 - 50        Residue name    resName       Residue name.
							if(line.length()>=51 )
								{
									String aa = line.substring(48-1, 51-1);
									if( aa != null)
										{
											aa = aa.trim();
											if( aa.length()>0)
												seq.append(get1LetterCodeFrom3LettersCode(aa));
										}
								}
							//52 - 54        Residue name    resName       Residue name.
							if(line.length()>=55 )
								{
									String aa = line.substring(52-1, 55-1);
									if( aa != null)
										{
											aa = aa.trim();
											if( aa.length()>0)
												seq.append(get1LetterCodeFrom3LettersCode(aa));
										}
								}
							//56 - 58        Residue name    resName       Residue name.
							if(line.length()>=59 )
								{
									String aa = line.substring(56-1, 59-1);
									if( aa != null)
										{
											aa = aa.trim();
											if( aa.length()>0)
												seq.append(get1LetterCodeFrom3LettersCode(aa));
										}
								}
							//60 - 62        Residue name    resName       Residue name.
							if(line.length()>=63 )
								{
									String aa = line.substring(60-1, 63-1);
									if( aa != null)
										{
											aa = aa.trim();
											if( aa.length()>0)
												seq.append(get1LetterCodeFrom3LettersCode(aa));
										}
								}
							//64 - 66        Residue name    resName       Residue name.
							if(line.length()>=67 )
								{
									String aa = line.substring(64-1, 67-1);
									if( aa != null)
										{
											aa = aa.trim();
											if( aa.length()>0)
												seq.append(get1LetterCodeFrom3LettersCode(aa));
										}
								}
							//68 - 70        Residue name    resName       Residue name.
							if(line.length()>=71 )
								{
									String aa = line.substring(68-1, 71-1);
									if( aa != null)
										{
											aa = aa.trim();
											if( aa.length()>0)
												seq.append(get1LetterCodeFrom3LettersCode(aa));
										}
								}
							retrievingChains.put(chainName.intern(), seq);
						
							//System.out.println("seq "+seq.toString());
							
							//StringBufferPool.returnString(seq);
						}
					
					//cleaning
					DataBufferPool.returnString(line);
					
					
						}
					reader.closeFile();
					//cleaning
					if( retrievingChains !=null)
						{
							if( chains == null)
								
								chains = new ArrayList<Chain>();
							for (Enumeration elements = retrievingChains.keys() ; elements.hasMoreElements();  )
								{
									String chainID = (String)elements.nextElement();
									if( chainID == null|| chainID.length()<1) continue;
									StringBuffer seq =  retrievingChains.get(chainID);
									if(seq == null || seq.length()<1) continue;
									chains.add(new PDBUtil.Chain(entry.getPdbID()+"_"+chainID, seq.toString()));
									
									StringBufferPool.returnString(seq);
								}
							retrievingChains.clear();
							retrievingChains = null;
						}
					
				}
			catch (IOException ioe)
				{
					logger.sessionError("error access to PDB file "+pdbFile.getName()+".",ioe);
				}
			if (!this.logger.isProduction()) logger.sessionDebug("retrieved chain for PDBID: "+entry.getPdbID()+" "+chains.get(0).getSequence());
			return chains;
		}
		
/**
 * Converts a three letter code of AminoAcid to the one letter code of AminoAcid.
 * @param threeLettersCode the three letter code to convert
 * @return the one letter code associated to the three letter code, if found. null otherwise
 */
		private static String get1LetterCodeFrom3LettersCode(String threeLettersCode)
			{
				/*System.out.println("threeLettersCode to convert |"+threeLettersCode+"|");*/
				if (threeLettersCode == null || threeLettersCode.length()!=3)
					{
						/*System.out.println("incorrect threeLettersCode "+threeLettersCode);*/
						if( threeLettersCode.length() ==1 ) return threeLettersCode;
						return null;
					}
				String oneLetterCode = getAminoAcids().get(threeLettersCode.toLowerCase());
				if(oneLetterCode == null || oneLetterCode.length()!=1 ) oneLetterCode ="X";
				return oneLetterCode;
			}
		
		/**
		 * Represents a chain contains in a PDB entry/file
		 * 
		 * @author <a href="mailto:pierre@ebi.ac.uk">Pierre MARGUERITE</a>
		 * @version 1
		 * @since 28 oct. 2005 
		 *
		 */
		public static class Chain
		{
			/**
			 * name of the chain
			 */
			private String name = null;
			private String sequence = null;
			
			/**
			 * 
			 * Constructor for chain 
			 *
			 * @param _name name of the chain
			 * @param _sequence	sequence of the chain
			 */
			public Chain (String _name, String _sequence)
				{
					this.name = _name;
					this.sequence = _sequence;
				}

			/**
			 * Retrieves the  name attribut
			 *
			 * @return Returns the name.
			 */
			public String getName()
				{
					return this.name;
				}

			/**
			 * Retrieves the  sequence attribut
			 *
			 * @return Returns the sequence.
			 */
			public String getSequence()
				{
					return this.sequence;
				}
			
			
		}

		/**
		 * Retrieves the  aminoAcids attribut
		 *
		 *AAs = Hashtable<String, String>();
				AAsAla	A	hydrophobic
				Arginine	Arg	R	free amino group makes it basic and hydrophilic
				Asparagine	Asn	N	carbohydrate can be covalently linked ("N-linked) to its -NH
				Aspartic acid	Asp	D	free carboxyl group makes it acidic and hydrophilic
				Cysteine	Cys	C	oxidation of their sulfhydryl (-SH) groups link 2 Cys (S-S)
				Glutamic acid	Glu	E	free carboxyl group makes it acidic and hydrophilic
				Glutamine	Gln	Q	moderately hydrophilic
				Glycine	Gly	G	so small it is amphiphilic (can exist in any surroundings)
				Histidine	His	H	basic and hydrophilic
				Isoleucine	Ile	I	hydrophobic
				Leucine	Leu	L	hydrophobic
				Lysine	Lys	K	strongly basic and hydrophilic
				Methionine	Met	M	hydrophobic
				Phenylalanine	Phe	F	very hydrophobic
				Proline	Pro	P	causes kinks in the chain
				Serine	Ser	S	carbohydrate can be covalently linked ("O-linked") to its -OH
				Threonine	Thr	T	carbohydrate can be covalently linked ("O-linked") to its -OH
				Tryptophan	Trp	W	scarce in most plant proteins
				Tyrosine	Tyr	Y	a phosphate or sulfate group can be covalently attached to its -OH
				Valine	Val	V
				
				A R D N B C E Z Q G H I L K M F P S T W Y V X
		 *
		 * @return Returns the aminoAcids.
		 */
		private static Hashtable<String, String> getAminoAcids()
			{
				if( AminoAcids == null)
					{//lazy initialisation
						AminoAcids = new Hashtable<String, String>();
						AminoAcids.put("Ala".toLowerCase(),	"A");
						AminoAcids.put("Arg".toLowerCase(), "R");
						AminoAcids.put("Asp".toLowerCase(), "D");	
						AminoAcids.put("Asn".toLowerCase(), "N");
						AminoAcids.put("Asx".toLowerCase(), "B");//       Aspartic acid or Asparagine
						AminoAcids.put("Cys".toLowerCase(), "C");
						AminoAcids.put("Glu".toLowerCase(), "E");
						AminoAcids.put("Glx".toLowerCase(), "Z");//       Glutamine or Glutamic acid.
						AminoAcids.put("Gln".toLowerCase(), "Q");
						AminoAcids.put("Gly".toLowerCase(), "G");
						AminoAcids.put("His".toLowerCase(), "H");
						AminoAcids.put("Ile".toLowerCase(), "I");
						AminoAcids.put("Leu".toLowerCase(), "L");
						AminoAcids.put("Lys".toLowerCase(), "K");
						AminoAcids.put("Met".toLowerCase(), "M");
						AminoAcids.put("Phe".toLowerCase(), "F");
						AminoAcids.put("Pro".toLowerCase(), "P");
						AminoAcids.put("Ser".toLowerCase(), "S");
						AminoAcids.put("Thr".toLowerCase(), "T");
						AminoAcids.put("Trp".toLowerCase(), "W");
						AminoAcids.put("Tyr".toLowerCase(), "Y");
						AminoAcids.put("Val".toLowerCase(), "V");
						AminoAcids.put("Xaa".toLowerCase(), "X");//       Any amino acid.
				       
				       
					}
				return AminoAcids;
			}
		
	}
